import json
from collections.abc import Callable
from typing import Any, TypeVar

from pydantic import BaseModel

TOOL_REFUSAL = json.dumps({"success": False, "reason": "User refused to allow tool"})

# Type variables for type safety
T = TypeVar("T", bound=BaseModel)

# Registry to map model names to (model class, handler function) pairs
_model_registry: dict[str, tuple[type[BaseModel], Callable[[Any], Any]]] = {}


def get_model_registry():
    return _model_registry


# Decorator to register models handlers
def register_model(handler_func: Callable[[T], Any]):
    # Get the type annotation of the model parameter
    model_class = handler_func.__annotations__.get("model")
    if not model_class:
        error_msg = (
            f"Handler {handler_func.__name__} must have a typed 'model' parameter"
        )

        raise ValueError(error_msg)

    registry = get_model_registry()
    registry[model_class.__name__] = (model_class, handler_func)
    return handler_func


# Function to process incoming requests
def process_tool_request(model_name: str, data: dict[str, Any]):
    registry = get_model_registry()
    if model_name not in registry:
        error_msg = f"Unknown tool: {model_name}"
        raise ValueError(error_msg)

    model_class, handler_func = registry[model_name]

    # Instantiate the model with the provided data
    model_instance = model_class(**data)

    # Call the handler with the model instance
    return handler_func(model_instance)



import sys

from pyaudio import PyAudio

from llm_chat_term.audio.pyaudio_no_log import PyAudioNoLog
from llm_chat_term.audio.speech_to_text import transcribe_speech
from llm_chat_term.config import config
from llm_chat_term.ui.chat_ui import ChatUI
from llm_chat_term.utils import get_api_key


def get_audio_input_devices(p: PyAudio) -> dict[str, int]:
    devices: dict[str, int] = {}
    for i in range(p.get_device_count()):
        device_info = p.get_device_info_by_index(i)
        max_input_channels = int(device_info["maxInputChannels"])
        name = str(device_info["name"])
        if max_input_channels > 0:
            devices[name] = i

    return devices


def get_user_device_idx() -> int:
    audio_device = config.audio_device
    with PyAudioNoLog() as pyaudio_initializer:
        p = pyaudio_initializer
        available_devices = get_audio_input_devices(p)
        p.terminate()

    if audio_device:
        found_audio_device = available_devices.get(audio_device)
        if found_audio_device:
            return found_audio_device

    # User hasn't selected audio device or we couldn't find it
    selected_audio_device = ChatUI.select_audio_device(available_devices)
    return available_devices[selected_audio_device]


def process_voice_command() -> str:
    openai_api_key = get_api_key("openai")
    if not openai_api_key:
        sys.stderr.write("API key for OpenAI not found.\n")
        sys.stderr.write("It is needed for transcribing audio.\n")
        return ""

    audio_device_idx = get_user_device_idx()
    captured_speech = transcribe_speech(
        audio_device_idx, openai_api_key.get_secret_value()
    )
    return captured_speech or ""



import mimetypes
import sys
from pathlib import Path

import trafilatura
from pydantic import HttpUrl

from llm_chat_term.exceptions import FileReadError, UrlReadError


def parse_insert_commands(user_input: str) -> str:
    result_lines: list[str] = []

    for line in user_input.splitlines():
        if line.strip().startswith(":read "):
            # Extract the file path
            file_path = Path(line.strip()[6:])
            file_path = file_path.expanduser()
            sys.stdout.write(f"Embedding file: {file_path}\n")

            # Check if file exists
            if not file_path.exists():
                error_msg = f"Error: File not found: {file_path}"
                raise FileReadError(error_msg)

            # Check if file is binary
            mime_type, _ = mimetypes.guess_type(file_path)
            if mime_type and not mime_type.startswith(("text/", "application/json")):
                error_msg = f"Error: :read file appears to be binary: {file_path}"
                raise FileReadError(error_msg)

            try:
                # Try to read the file
                with file_path.open(encoding="utf-8") as f:
                    file_contents = f.read()
                # Add file contents instead of the :read line
                result_lines.append(file_contents.rstrip())
            except UnicodeDecodeError as e:
                error_msg = f"Error: :read file appears to be binary: {file_path}"
                raise FileReadError(error_msg) from e
            except Exception as e:
                error_msg = f"Error: Could not :read file: {file_path}: {e!s}"
                raise FileReadError(error_msg) from e
        elif line.strip().startswith(":web "):
            url = line.strip()[5:]
            sys.stdout.write(f"Embedding url: {url}\n")
            try:
                url = HttpUrl(url)
            except ValueError as e:
                error_msg = f"Error: Could not parse url {url}"
                raise UrlReadError(error_msg) from e

            try:
                downloaded = trafilatura.fetch_url(str(url))
                error_msg = f"Error: Could not :web url: {url}"
                if downloaded:
                    text = trafilatura.extract(
                        downloaded,
                        with_metadata=True,
                        deduplicate=True,
                    )
                    if not text:
                        raise UrlReadError(error_msg)
                    result_lines.append(
                        f"The following text was extracted from {url}:\n{text}"
                    )
                    continue

                raise UrlReadError(error_msg)
            except Exception as e:
                error_msg = f"Error: Could not :web url: {url}: {e!s}"
                raise UrlReadError(error_msg) from e

        else:
            # Keep the original line
            result_lines.append(line)

    return "\n".join(result_lines)



"""Terminal UI for the LLM chatbot using prompt_toolkit and rich."""

from typing import cast, override

from langchain_core.messages import AIMessage, BaseMessage, HumanMessage
from prompt_toolkit import PromptSession
from prompt_toolkit.history import InMemoryHistory
from prompt_toolkit.styles import Style
from rich.console import Console, ConsoleOptions, Group, RenderResult
from rich.live import Live
from rich.markdown import CodeBlock, Markdown
from rich.rule import Rule
from rich.syntax import Syntax
from rich.text import Text

from llm_chat_term.config import config
from llm_chat_term.ui.audio_device_selector import select_audio_device
from llm_chat_term.ui.chat_selector import create_new_chat, select_chat
from llm_chat_term.ui.confirm_prompt import confirm_prompt
from llm_chat_term.ui.help import print_help
from llm_chat_term.ui.model_selector import select_model


class CodeBlockNoPadding(CodeBlock):
    """A code block with syntax highlighting."""

    @override
    def __rich_console__(
        self, console: Console, options: ConsoleOptions
    ) -> RenderResult:
        code = str(self.text).rstrip()
        syntax = Syntax(
            code, self.lexer_name, theme=self.theme, word_wrap=True, padding=0
        )
        yield syntax


class ChatUI:
    """Terminal UI for the chatbot."""

    def __init__(self):
        self.console = Console()

        # Set up the prompt style
        self.style = Style.from_dict(
            {
                "prompt": f"fg:{config.colors.user}",
            }
        )

        # Create a session for the prompt
        self.session = PromptSession[str](
            style=self.style,
            history=InMemoryHistory(),
            multiline=True,
            vi_mode=False,
            prompt_continuation=lambda width, line_number, is_soft_wrap: " "
            * len(config.ui.prompt_symbol),
        )

        # Track whether we're currently streaming a response
        self.streaming = False
        # Track whether we're in a thinking block
        self.thinking = False
        self.current_response = ""
        self.live = Live()

    def _get_ai_title(self):
        content = Text(
            f" {config.ui.assistant} ", style=f"bold white on {config.colors.assistant}"
        )
        return Rule(content, style=f"on {config.colors.assistant}", characters=" ")

    def _get_user_title(self, model_name: str = "", chat_id: str = ""):
        if not model_name:
            model_name = ""
            chat_name = ""
        else:
            model_name = f"[{model_name}] "
            chat_name = f" [{chat_id}]" if chat_id else " [Anonymous chat]"

        width = self.console.width
        left_width = width // 3
        center_width = width // 3
        right_width = width - left_width - center_width

        formatted = (
            f"{chat_name:<{left_width}}"
            f"{config.ui.user:^{center_width}}"
            f"{model_name:>{right_width}}"
        )

        return Text(formatted, style=f"bold black on {config.colors.user}")

    def _get_markdown(self, content: str):
        try:
            text = Markdown(content)
            text.elements["fence"] = CodeBlockNoPadding
            text.elements["code_block"] = CodeBlockNoPadding
        except Exception:
            text = Text(content)

        return text

    def display_welcome_message(self):
        """Display a welcome message when a chat starts."""
        self.console.print(
            "Type your messages and press Alt(Esc)+Enter to send. Use Enter for newlines.",
            style=config.colors.system,
        )
        self.console.print(
            "Type '/help' for commands help, '/exit' or Ctrl+D to exit.",
            style=config.colors.system,
        )
        self.console.print()

    def display_help(self):
        print_help(self.console)

    def display_info(self, chat_id: str):
        if chat_id:
            self.console.print(
                f"Selected chat: {chat_id}", style=f"bold {config.colors.system}"
            )
        else:
            self.console.print("Anonymous chat", style=f"bold {config.colors.system}")

    def display_loader(self):
        self.console.print("[yellow]Contemplating...[/yellow]")

    @staticmethod
    def create_new_chat(*, allow_blank: bool = False):
        return create_new_chat(allow_blank=allow_blank)

    @staticmethod
    def select_chat():
        return select_chat()

    @staticmethod
    def select_model():
        return select_model()

    @staticmethod
    def select_audio_device(available_devices: dict[str, int]):
        return select_audio_device(available_devices)

    @staticmethod
    def display_prompt(question: str) -> bool:
        return confirm_prompt(question)

    def get_user_input(
        self, model_name: str, chat_id: str, *, prefilled_prompt: str = ""
    ) -> str:
        """Get multiline input from the user."""
        try:
            self.console.print(self._get_user_title(model_name, chat_id))
            user_input = self.session.prompt(
                config.ui.prompt_symbol,
                style=self.style,
                default=prefilled_prompt,
            )
            return user_input.strip()
        except KeyboardInterrupt:
            # Handle Ctrl+C
            return "exit"
        except EOFError:
            # Handle Ctrl+D
            return "exit"

    def render_conversation(self, messages: list[BaseMessage], chat_id: str):
        self.console.clear()
        # Display welcome message
        self.display_welcome_message()
        chat_name = chat_id if chat_id else "Anonymous chat"

        content = Text(
            f" CONVERSATION START [{chat_name}]",
            style=f"bold black on {config.colors.system}",
        )
        rule = Rule(content, style=f"on {config.colors.system}", characters=" ")

        self.console.print(rule)
        for message in messages:
            message.content = cast(str, message.content)
            if isinstance(message, HumanMessage):
                self.console.print(self._get_user_title())
                self.console.print(
                    f"[{config.colors.user}]{config.ui.prompt_symbol}[/] {message.content}"
                )
            elif isinstance(message, AIMessage):
                self.console.print(self._get_ai_title())
                self.console.print(self._get_markdown(message.content))
                self.console.print()
            else:
                continue

    def _update_live(self):
        text = self._get_markdown(self.current_response)

        content = Group(self._get_ai_title(), text)
        self.live.update(content, refresh=True)

    def stream_token(self, token: str, chunk_type: str):
        """Display a streaming token from the assistant."""
        if not self.streaming:
            self.console.clear()
            self.live.start()
            self.streaming = True
        if chunk_type == "thinking" and not self.thinking:
            self.thinking = True
            self.current_response += "\\<think>\n"
        elif chunk_type == "text" and self.thinking:
            # Here starts the actual response, clear the thinking part
            self.thinking = False
            self.current_response = ""
        elif chunk_type == "prompt_tool" and self.streaming:
            self.live.stop()
            self.streaming = False

        self.current_response += token
        self._update_live()

    def end_streaming(self):
        """End the streaming response and print a newline."""
        if self.streaming:
            self.live.stop()
            self.console.print()
            self.streaming = False
            self.current_response = ""



import sys

from langchain_core.messages import SystemMessage
from pydantic import SecretStr

from llm_chat_term import db, utils
from llm_chat_term.audio.audio_entrypoint import handle_voice
from llm_chat_term.config import ModelConfig, config
from llm_chat_term.llm.insert_commands import parse_insert_commands
from llm_chat_term.llm.llm_client import LLMClient
from llm_chat_term.ui.chat_ui import ChatUI


class LLMChat:
    def __init__(self):
        self.ui = ChatUI()
        available_model: ModelConfig | None = None

        self.api_key = SecretStr("")
        for model in config.llm.models:
            try:
                self.api_key = utils.get_api_key(model.provider)
            except ValueError:
                continue
            else:
                available_model = model
                break

        if available_model is None:
            error_msg = (
                "Could not find an API key for any of the providers.\n"
                f"Please add them in {db.get_config_file()} or as an env var\n"
            )
            sys.stderr.write(error_msg)
            sys.exit(1)
        self.model = available_model
        self.chat_id = self.initialize()
        self.client = LLMClient(self.model, self.api_key)
        if self.chat_id:
            self.client.parse_messages(self.chat_id)
            self.ui.render_conversation(self.client.messages, self.chat_id)
        else:
            self.client.messages = [SystemMessage(config.llm.system_prompt)]
            self.ui.render_conversation(self.client.messages, self.chat_id)

    def initialize(self) -> str:
        try:
            chat_id = self.ui.select_chat()
        except Exception as e:
            error_msg = f"Error: {e!s}\n"
            sys.stderr.write(error_msg)
            sys.exit(1)
        return chat_id

    def start_chat(self):
        should_think = False
        recorded_prompt: str = ""

        while True:
            # Get user input
            user_input = self.ui.get_user_input(
                self.model.name, self.chat_id, prefilled_prompt=recorded_prompt
            )
            recorded_prompt = ""
            # Skip empty inputs
            if not user_input:
                continue
            # Check for commands
            if user_input in (":exit", "exit"):
                sys.stderr.write("\nExiting...\n")
                break
            if user_input == ":help":
                self.ui.display_help()
                continue
            if user_input == ":info":
                self.ui.display_info(self.chat_id)
                continue
            if user_input in [":edit", ":e"]:
                if self.chat_id:
                    utils.open_in_editor(self.chat_id)
                    self.client.parse_messages(self.chat_id)
                    self.ui.render_conversation(self.client.messages, self.chat_id)
                else:
                    new_chat_id = self.ui.create_new_chat(allow_blank=False)
                    if new_chat_id:
                        messages_dict = self.client.get_conversation_history()
                        db.save_chat_history(new_chat_id, messages_dict)
                        utils.open_in_editor(new_chat_id)
                        self.chat_id = new_chat_id
                        self.client.parse_messages(self.chat_id)
                        self.ui.render_conversation(self.client.messages, self.chat_id)
                    else:
                        sys.stderr.write(
                            "Cannot edit conversation without a chat name...\n"
                        )
                continue
            if user_input == ":model":
                model = self.ui.select_model()
                try:
                    self.api_key = utils.get_api_key(model.provider)
                except ValueError as e:
                    error_msg = f"Error: {e!s}\n"
                    sys.stderr.write(error_msg)
                else:
                    self.model = model
                    self.client.configure_model(self.model, self.api_key)
                continue
            if user_input == ":chat":
                self.chat_id = self.ui.select_chat()
                self.client.parse_messages(self.chat_id)
                self.ui.render_conversation(self.client.messages, self.chat_id)
                continue
            if user_input == ":redraw":
                self.client.parse_messages(self.chat_id)
                self.ui.render_conversation(self.client.messages, self.chat_id)
                continue
            if user_input == ":agent on":
                self.client.agent_mode = True
                self.ui.console.print("Agent mode enabled", style="bold green")
                continue
            if user_input == ":agent off":
                self.client.agent_mode = False
                self.ui.console.print("Agent mode disabled", style="bold orchid")
                continue
            if user_input.startswith(":think"):
                should_think = True
            elif user_input.startswith(":v"):
                recorded_prompt = handle_voice()
                continue
            else:
                try:
                    user_input = parse_insert_commands(user_input)
                except Exception as e:
                    error_msg = f"Error: {e!s}\n"
                    sys.stderr.write(error_msg)
                    continue

            # Get and display streaming response
            self.ui.display_loader()
            try:
                self.client.get_response(
                    user_input,
                    self.ui.stream_token,
                    chat_id=self.chat_id,
                    should_think=should_think,
                )
            except Exception as e:
                error_msg = f"Something went wrong... {e!s}\n"
                sys.stderr.write(error_msg)
            should_think = False

            # End streaming
            self.ui.end_streaming()

        return sys.exit(0)



import sys
import tempfile
import warnings
import wave
from pathlib import Path

import pyaudio
from pydub import AudioSegment

from llm_chat_term.audio.pyaudio_no_log import PyAudioNoLog

# Silence tye pydub warnings coming from broken regexes
warnings.filterwarnings("ignore", category=SyntaxWarning, module="pydub.utils")


def record_audio(
    audio_device_idx: int,
    sample_rate: int = 16000,
    channels: int = 1,
    chunk: int = 1024,
) -> str | None:
    """Record audio from microphone until Ctrl+C and save as MP3"""
    # Initialize PyAudio, capturing stdout and stderr to avoid it
    # flooding the console with warnings
    with PyAudioNoLog() as pyaudio_initializer:
        p = pyaudio_initializer

    with tempfile.NamedTemporaryFile(suffix=".wav", delete=False) as temp:
        temp_wav = temp.name
    with tempfile.NamedTemporaryFile(suffix=".mp3", delete=False) as temp:
        temp_mp3 = temp.name

    sys.stdout.write("Recording... Press Ctrl+C to stop")
    sys.stdout.flush()
    try:
        # Open stream
        stream = p.open(
            format=pyaudio.paInt16,
            channels=channels,
            rate=sample_rate,
            input=True,
            input_device_index=audio_device_idx,
            frames_per_buffer=chunk,
        )
    except Exception as e:
        sys.stderr.write(f"Error capturing audio: {e}\n")
        return None

    frames: list[bytes] = []

    try:
        while True:
            data = stream.read(chunk)
            frames.append(data)
    except KeyboardInterrupt:
        sys.stdout.write("\033[2K")  # Clear the entire line and return to start
        sys.stdout.flush()
        sys.stdout.write("Recording stopped. Transcribing audio...\n")
    finally:
        # Stop and close the stream
        stream.stop_stream()
        stream.close()
        p.terminate()

    if frames:
        # Save as WAV first (temporary)
        wf = wave.open(temp_wav, "wb")  # noqa: SIM115
        wf.setnchannels(channels)
        wf.setsampwidth(p.get_sample_size(pyaudio.paInt16))
        wf.setframerate(sample_rate)
        wf.writeframes(b"".join(frames))
        wf.close()

        # Convert to MP3
        sound = AudioSegment.from_wav(temp_wav)
        sound.export(
            temp_mp3,
            format="mp3",
            bitrate="64k",
            parameters=["-ac", "1", "-ar", "16000"],
        )

        # Remove temporary WAV file
        Path(temp_wav).unlink()
        return temp_mp3

    sys.stderr.write("Failed to capture audio...\n")
    return None



from llm_chat_term.config import config, save_config
from llm_chat_term.ui.prompt_menu import Menu


def update_config(selected_audio_device: str) -> None:
    config.audio_device = selected_audio_device
    save_config(config)


def select_audio_device(audio_devices: dict[str, int]) -> str:
    options = list(audio_devices)
    menu = Menu(
        options,
        (
            " Select an audio device to record from "
            "(j/k to move, Enter to select, e to edit, d to delete, q to quit):\n"
        ),
    )

    result = menu.run()

    selected_audio_device = options[result]
    update_config(selected_audio_device)

    return selected_audio_device



import os
import subprocess

from pydantic import SecretStr

from llm_chat_term import db
from llm_chat_term.config import config


def get_api_key(provider: str) -> SecretStr:
    api_key: SecretStr = next(
        (x.api_key for x in config.llm.api_keys if x.provider == provider),
        SecretStr(""),
    )
    if not api_key.get_secret_value():
        from_env = os.getenv(f"{provider.upper()}_API_KEY", None)
        if not from_env:
            error_msg = (
                f"API key for provider {provider.upper()} is not set. Please set it in your "
                f"{db.get_config_file()} file or as an env var."
            )
            raise ValueError(error_msg)
        api_key = SecretStr(from_env)

    return api_key


def delete_chat(chat_id: str) -> None:
    full_path = db.get_chat_file(chat_id)
    full_path.unlink()


def open_in_editor(chat_id: str):
    full_path = db.get_chat_file(chat_id)
    editor = os.environ.get("EDITOR", "vim")
    # Assume that the user knows what EDITOR is set
    subprocess.call([editor, str(full_path)])  # noqa: S603


def has_audio_support():
    try:
        import pyaudio  # pyright: ignore[reportUnusedImport]  # noqa: F401
        import pydub  # pyright: ignore[reportUnusedImport]  # noqa: F401

    except ImportError:
        return False
    else:
        return True



"""Configuration module for the terminal LLM chatbot."""

import sys

import yaml
from pydantic import BaseModel, Field, SecretStr

from llm_chat_term.db import get_config_file


class ModelConfig(BaseModel):
    provider: str
    name: str
    temperature: float | None = None


def get_default_models():
    return [
        ModelConfig(
            provider="anthropic",
            name="claude-3-7-sonnet-20250219",
        ),
        ModelConfig(provider="openai", name="gpt-4o"),
        ModelConfig(provider="openai", name="o3-mini"),
    ]


class ApiKey(BaseModel):
    provider: str
    api_key: SecretStr = Field(default=SecretStr(""))


def get_default_api_keys() -> list[ApiKey]:
    return [ApiKey(provider="anthropic"), ApiKey(provider="openai")]


class LLMConfig(BaseModel):
    models: list[ModelConfig] = Field(default_factory=get_default_models)
    api_keys: list[ApiKey] = Field(default_factory=get_default_api_keys)
    system_prompt: str = (
        "You are a helpful assistant responding to a user's questions in a PC terminal application.\n"
        "The user is an experienced software engineer, your answers should be concise and not repetitive.\n"
        "Skip conclusions and summarizations of your answers.\n"
        "If the user asks for a change in code, don't return the whole code, just the changed segment(s).\n"
        "Return your answers in markdown format, and wrap code in ``` blocks, but avoid using headings."
    )


class UIConfig(BaseModel):
    prompt_symbol: str = ">>> "
    user: str = "user"
    assistant: str = "assistant"


class ColorConfig(BaseModel):
    user: str = "cyan"
    assistant: str = "grey39"
    system: str = "yellow"


class AppConfig(BaseModel):
    llm: LLMConfig = Field(default_factory=LLMConfig)
    ui: UIConfig = Field(default_factory=UIConfig)
    colors: ColorConfig = Field(default_factory=ColorConfig)
    audio_device: str = ""


def save_config(conf: AppConfig) -> None:
    """Save the config model to the YAML file."""
    config_file = get_config_file()
    conf_dict = conf.model_dump()

    for api_key in conf_dict["llm"]["api_keys"]:
        api_key["api_key"] = api_key["api_key"].get_secret_value()

    with config_file.open("w") as f:
        yaml.dump(conf_dict, f, default_flow_style=False, sort_keys=False)


def load_config() -> AppConfig:
    """Load configuration from YAML file with fallbacks to default values."""
    config_file = get_config_file()

    # Start with default configuration
    config = AppConfig()

    # Load configuration from file if it exists
    if not config_file.exists():
        try:
            save_config(config)
            sys.stdout.write(f"Created default configuration file at {config_file}\n")
        except Exception as e:
            error_msg = f"Error creating default config file: {e}\n"
            sys.stderr.write(error_msg)
            sys.exit(1)
    else:
        try:
            with config_file.open() as f:
                user_config = yaml.safe_load(f)

            return AppConfig.model_validate(user_config)
        except Exception as e:
            error_msg = f"Error loading config file: {e!s}\n"
            sys.stderr.write(error_msg)
            sys.exit(1)

    return config


config: AppConfig = load_config()



from prompt_toolkit import prompt
from prompt_toolkit.key_binding import KeyBindings, KeyPressEvent


def confirm_prompt(message: str) -> bool:
    """
    Display a yes/no prompt that immediately returns on y/n keypress.
    Returns True for 'y' and False for 'n'.
    """
    bindings = KeyBindings()
    result = False

    @bindings.add("y")
    @bindings.add("Y")
    def _(event: KeyPressEvent):
        nonlocal result
        result = True
        event.app.exit()

    @bindings.add("n")
    @bindings.add("N")
    def _(event: KeyPressEvent):
        nonlocal result
        result = False
        event.app.exit()

    try:
        prompt(
            message + " (y/n) ",
            key_bindings=bindings,
            default="",
        )
    except KeyboardInterrupt:
        return False

    return result



"""Main application module for the terminal LLM chatbot."""

from typing import NoReturn

from llm_chat_term.llm.llm_chat import LLMChat


def main() -> NoReturn:
    """Run the terminal LLM chatbot application."""

    chat = LLMChat()
    chat.start_chat()


if __name__ == "__main__":
    main()



from pydantic import BaseModel, Field


class CatFileCommand(BaseModel):
    """Execute the `cat` command with arbitrary arguments.
    The arguments will be passed through shlex.split().
    Use it to get the contents of one or more files in the current directory.
    Returns a success boolean, the return code, stdout, and stderr of the command as json.
    Example:
    To execute `cat file.txt` call this tool with {"arguments": "file.txt"}
    """

    arguments: str = Field(..., description="The arguments for the cat command")


class GitCommand(BaseModel):
    """Execute the `git` command with arbitrary arguments.
    The arguments will be passed through shlex.split().
    Use it to execute git related commands (e.g. status, diff, add, commit, branch, rebase).
    If some action is destructive(not reversible, e.g. "git checkout file"), don't execute it.
    Returns a success boolean, the return code, stdout, and stderr of the command as json.
    Example:
    To execute `git status -s` call this tool with {"arguments": "status -s"}
    """

    arguments: str = Field(..., description="The arguments for the git command")


# These will be bound to the LLM model
tools = [
    CatFileCommand,
    GitCommand,
]



import os
from pathlib import Path
from urllib.parse import quote_plus, unquote_plus

MESSAGE_INDICATOR = "▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒▒"


def get_chat_file(chat_id: str) -> Path:
    return _get_chats_dir() / _encode_filename(chat_id)


def get_config_file() -> Path:
    return _get_config_dir() / "config.yaml"


def _get_config_dir() -> Path:
    """Get platform-specific config directory for llm_chat_term."""
    home = Path.home()

    if os.name == "nt":  # Windows
        config_dir = home / "AppData" / "Local" / "llm_chat_term"
    elif os.name == "posix":  # Linux, macOS, etc.
        xdg_config_home = os.environ.get("XDG_CONFIG_HOME")
        if xdg_config_home:
            config_dir = Path(xdg_config_home) / "llm_chat_term"
        elif (home / ".config").exists():
            config_dir = home / ".config" / "llm_chat_term"
        else:  # macOS fallback
            config_dir = home / "Library" / "Preferences" / "llm_chat_term"
    else:
        config_dir = home / ".llm_chat_term"  # Fallback

    config_dir.mkdir(parents=True, exist_ok=True)
    return config_dir


def _get_data_dir() -> Path:
    """Get platform-specific data directory for llm_chat_term."""
    home = Path.home()

    if os.name == "nt":  # Windows
        data_dir = home / "AppData" / "Local" / "llm_chat_term"
    elif os.name == "posix":  # Linux, macOS, etc.
        xdg_data_home = os.environ.get("XDG_DATA_HOME")
        if xdg_data_home:
            data_dir = Path(xdg_data_home) / "llm_chat_term"
        elif (home / ".local" / "share").exists():  # Linux
            data_dir = home / ".local" / "share" / "llm_chat_term"
        else:  # macOS
            data_dir = home / "Library" / "Application Support" / "llm_chat_term"
    else:
        data_dir = home / ".llm_chat_term"  # Fallback

    data_dir.mkdir(parents=True, exist_ok=True)
    return data_dir


def _get_chats_dir() -> Path:
    """Get the directory where the chats are saved"""
    chats_dir = _get_data_dir() / "chats"
    chats_dir.mkdir(parents=True, exist_ok=True)
    return chats_dir


def _encode_filename(chat_id: str) -> str:
    """Encode chat ID to a safe filename."""
    return f"{quote_plus(chat_id)}.txt"


def _decode_filepath(file_path: Path) -> str:
    """Decode filename back to original chat ID."""
    # Remove .txt extension and decode
    return unquote_plus(file_path.stem)


# TODO: Sometimes we get a ".txt" file in the chats dir
# Investigate
def save_chat_history(chat_id: str, messages: list[dict[str, str]]):
    """Save chat history to a text file.

    Args:
        chat_id: Identifier for the chat session
        messages: List of message dictionaries
    """
    file_path = get_chat_file(chat_id)

    with file_path.open("w", encoding="utf-8") as f:
        max_role_width = max([len(message["role"]) for message in messages])
        for message in messages:
            role = message["role"]
            padding = max_role_width - len(role)
            left_padding = padding // 2
            right_padding = padding - left_padding
            padded_role = " " * left_padding + role + " " * right_padding

            f.write(f"{MESSAGE_INDICATOR} {padded_role} {MESSAGE_INDICATOR}\n")
            f.write(f"{message['content']}\n")


def load_chat_history(chat_id: str) -> list[dict[str, str]]:
    """Load chat history from a text file.

    Args:
        chat_id: Identifier for the chat session

    Returns:
        List of message dictionaries
    """
    file_path = get_chat_file(chat_id)

    if not file_path.exists():
        return []

    messages: list[dict[str, str]] = []
    with file_path.open(encoding="utf-8") as f:
        lines = f.read().splitlines()
        last_index = len(lines) - 1

        msg_role = ""
        msg_content = ""
        for i, line in enumerate(lines):
            if line.startswith(f"{MESSAGE_INDICATOR}") and line.endswith(
                f"{MESSAGE_INDICATOR}"
            ):
                messages.append(
                    {
                        "role": msg_role,
                        "content": msg_content.rstrip(),
                    }
                )
                msg_content = ""
                # Extract next message type if not last file line
                msg_role = line.split(f"{MESSAGE_INDICATOR}")[1].strip()
            elif i == last_index:
                msg_content += line + "\n"
                messages.append(
                    {
                        "role": msg_role,
                        "content": msg_content.rstrip(),
                    }
                )
            else:
                msg_content += line + "\n"

    return messages


def list_all_chats() -> list[str]:
    """List all available chat histories.

    Returns:
        List of all chat_ids
    """
    data_dir = _get_chats_dir()
    chats_with_time: list[tuple[str, float]] = []

    for file_path in data_dir.glob("*.txt"):
        try:
            chat_id = _decode_filepath(file_path)
            mod_time = file_path.stat().st_mtime
            chats_with_time.append((chat_id, mod_time))
        except Exception:
            # Skip files that can't be decoded properly
            continue

    chats_with_time.sort(key=lambda x: x[1], reverse=True)

    return [chat_id for chat_id, _ in chats_with_time]



import sys
from pathlib import Path

from openai import OpenAI

from llm_chat_term.audio.recorder import record_audio


def transcribe_speech(audio_device_idx: int, openai_api_key: str) -> str | None:
    mp3_file = record_audio(audio_device_idx)
    if not mp3_file:
        return None
    mp3_file = Path(mp3_file)

    client = OpenAI(api_key=openai_api_key)
    try:
        audio_file = mp3_file.open("rb")
        transcription = client.audio.transcriptions.create(
            model="whisper-1", file=audio_file, language="en"
        )
        audio_file.close()
    except Exception as e:
        sys.stderr.write(f"Error transcribing audio: {e}\n")
        return None
    finally:
        mp3_file.unlink()

    result = transcription.text
    if result == "Thanks for watching!":
        return None
    return result



import sys

from llm_chat_term.utils import has_audio_support


def handle_voice():
    if not has_audio_support():
        sys.stderr.write("No voice support found. Install llm_chat_term[voice]\n")
        return ""

    from llm_chat_term.audio.voice_command import process_voice_command

    return process_voice_command()



import os

from pyaudio import PyAudio


class PyAudioNoLog:
    """
    PyAudio is noisy af every time you initialise it, which makes reading the
    log output rather difficult.  The output appears to be being made by the
    C internals, so I can't even redirect the logs with Python's logging
    facility.  Therefore the nuclear option was selected: swallow all stderr
    and stdout for the duration of PyAudio's use.

    Lifted and adapted from StackOverflow:
      https://stackoverflow.com/questions/11130156/
    """

    def __init__(self):
        # Open a pair of null files
        self.null_fds = [os.open(os.devnull, os.O_RDWR) for _ in range(2)]

        # Save the actual stdout (1) and stderr (2) file descriptors.
        self.save_fds = [os.dup(1), os.dup(2)]

        self.pyaudio = None

    def __enter__(self) -> PyAudio:
        # Assign the null pointers to stdout and stderr.
        os.dup2(self.null_fds[0], 1)
        os.dup2(self.null_fds[1], 2)

        self.pyaudio = PyAudio()

        return self.pyaudio

    def __exit__(self, *_):
        # Re-assign the real stdout/stderr back to (1) and (2)
        os.dup2(self.save_fds[0], 1)
        os.dup2(self.save_fds[1], 2)

        # Close all file descriptors
        for fd in self.null_fds + self.save_fds:
            os.close(fd)



from llm_chat_term.config import ModelConfig, config
from llm_chat_term.ui.prompt_menu import Menu


def select_model() -> ModelConfig:
    models = config.llm.models

    menu = Menu(
        [model.name for model in models],
        " Select a model (j/k to move, Enter to select, e to edit, d to delete, q to quit):\n",
    )

    result = menu.run()

    return models[result]



"""LLM client for the terminal chatbot using LangChain."""

import json
from collections.abc import Callable
from typing import cast

from langchain_anthropic import ChatAnthropic
from langchain_core.language_models import BaseChatModel
from langchain_core.messages import (
    AIMessage,
    BaseMessage,
    BaseMessageChunk,
    HumanMessage,
    SystemMessage,
    ToolMessage,
    ToolMessageChunk,
    message_chunk_to_message,
)
from langchain_openai import ChatOpenAI
from pydantic import SecretStr

from llm_chat_term import db
from llm_chat_term.config import ModelConfig, config
from llm_chat_term.llm.tools.definitions import tools
from llm_chat_term.llm.tools.main import TOOL_REFUSAL, process_tool_request
from llm_chat_term.ui.chat_ui import ChatUI


def get_chunk_text_and_type(chunk: BaseMessageChunk) -> tuple[str, str]:
    """Get the text content and type of the message."""
    if isinstance(chunk.content, str):
        return chunk.content, "text"

    # must be a list, extract the type from the first block
    content: list[dict[str, str] | str] = chunk.content
    first_block = content[0]
    if isinstance(first_block, str):
        chunk_type = "text"
    else:
        chunk_type: str = first_block.get("type", "text")
    return "".join(
        block if isinstance(block, str) else block.get(chunk_type, "")
        for block in chunk.content  # pyright: ignore[reportUnknownVariableType]
    ), chunk_type


class LLMClient:
    """Client for interacting with the LLM."""

    agent_mode = False

    def __init__(self, model: ModelConfig, api_key: SecretStr):
        """Initialize the LLM client with the configured model."""
        self.messages: list[BaseMessage] = []
        self.configure_model(model, api_key)

    def configure_model(self, model_config: ModelConfig, api_key: SecretStr) -> None:
        self.model_config = model_config
        if model_config.provider == "openai" and model_config.name.startswith("o"):
            temperature = 1.0
        else:
            temperature = model_config.temperature or 0.4
        temperature = 1.0
        if model_config.provider == "anthropic":
            self.model = ChatAnthropic(  # pyright: ignore[reportCallIssue]
                api_key=api_key,
                model=model_config.name,  # pyright: ignore[reportCallIssue]
                temperature=temperature,
                max_tokens=16384,  # pyright: ignore[reportCallIssue]
                stream_usage=False,
                streaming=True,
            )
            self.thinking_model = ChatAnthropic(  # pyright: ignore[reportCallIssue]
                api_key=api_key,
                model=model_config.name,  # pyright: ignore[reportCallIssue]
                temperature=1.0,  # Needs to be 1 for thinking
                max_tokens=16384,  # pyright: ignore[reportCallIssue]
                thinking={"type": "enabled", "budget_tokens": 2048},
                stream_usage=False,
                streaming=True,
            )
        elif model_config.provider == "openai":
            self.model = ChatOpenAI(
                api_key=api_key,
                model=model_config.name,
                temperature=temperature,
                max_tokens=16384,  # pyright: ignore[reportCallIssue]
                streaming=True,
            )
            self.thinking_model = self.model

    def add_user_message(self, content: str) -> None:
        """Add a user message to the conversation history."""
        self.messages.append(HumanMessage(content))

    def add_assistant_message(self, content: str) -> None:
        """Add an assistant message to the conversation history."""
        self.messages.append(AIMessage(content))

    def get_response(
        self,
        user_message: str,
        stream_callback: Callable[[str, str], None],
        *,
        chat_id: str = "",
        should_think: bool = False,
        # For :tmp handling, we don't append user message to conversation history
        # but we only send it for the current response
    ) -> None:
        """Get a response from the LLM and stream it through the callback."""
        if self.agent_mode:
            model = self.model.bind_tools(tools)
        else:
            # Thinking model only for claude, o3-mini always thinks, 4o never
            model = self.thinking_model if should_think else self.model
        model = cast(BaseChatModel, model)

        response = ""
        if user_message:
            self.messages.append(HumanMessage(user_message))
        is_tool = False
        tool_json = ""
        tool_name = ""
        tool_call_id = ""
        tool_message: BaseMessageChunk | None = None
        # TODO: o3-mini doesn't know what to do with response ToolMessage
        # Ditch langchain
        for chunk in model.stream(self.messages):
            if (
                hasattr(chunk, "tool_call_chunks")
                and isinstance(chunk.tool_call_chunks, list)
                and len(chunk.tool_call_chunks) > 0
                and isinstance(chunk.tool_call_chunks[0], dict)
            ):
                chunk = cast(ToolMessageChunk, chunk)
                block = cast(dict[str, str], chunk.tool_call_chunks[0])
                if not is_tool:
                    is_tool = True
                    tool_name = block["name"]
                    tool_call_id = block["id"]
                    tool_message = chunk
                else:
                    tool_json += block["args"]
                    if tool_message:
                        tool_message = tool_message + chunk
            else:
                text, chunk_type = get_chunk_text_and_type(chunk)
                stream_callback(text, chunk_type)
                response += chunk.text()
        if is_tool and tool_message and tool_call_id:
            ai_tool_message = message_chunk_to_message(tool_message)
            # Pause streaming to display the confirm prompt
            stream_callback("", "prompt_tool")
            confirm = ChatUI.display_prompt(f"Use tool {tool_name} with {tool_json}:")
            if confirm:
                stream_callback(
                    (f"\n\n*-- Calling tool {tool_name} with {tool_json}...*\n"),
                    "text",
                )
                tool_result = process_tool_request(tool_name, json.loads(tool_json))
                stream_callback(
                    f"{'**success**' if tool_result['success'] else '**failure**'}\n\n",
                    "text",
                )
                self.messages.append(ai_tool_message)
                self.messages.append(
                    ToolMessage(json.dumps(tool_result), tool_call_id=tool_call_id)
                )
                self.get_response("", stream_callback, chat_id=chat_id)
            else:
                stream_callback((f"\n\n-- Will not call tool {tool_name}.\n\n"), "text")
                self.messages.append(ai_tool_message)
                self.messages.append(
                    ToolMessage(TOOL_REFUSAL, tool_call_id=tool_call_id)
                )

        self.messages.append(AIMessage(response))
        if chat_id:
            db.save_chat_history(chat_id, self.get_conversation_history())

    def parse_messages(self, chat_id: str):
        messages_dict = db.load_chat_history(chat_id)
        self.messages = []
        for message in messages_dict:
            if message["role"] == "system":
                self.messages.append(SystemMessage(message["content"]))
            elif message["role"] == "user":
                self.messages.append(HumanMessage(message["content"]))
            elif message["role"] == "assistant":
                self.messages.append(AIMessage(message["content"]))

        if not any(isinstance(message, SystemMessage) for message in self.messages):
            self.messages.insert(0, SystemMessage(config.llm.system_prompt))
            db.save_chat_history(chat_id, self.get_conversation_history())

    def get_conversation_history(self) -> list[dict[str, str]]:
        """Get the conversation history as a list of dictionaries."""
        history: list[dict[str, str]] = []

        for message in self.messages:
            if isinstance(message, SystemMessage):
                role = "system"
            elif isinstance(message, HumanMessage):
                role = "user"
            elif isinstance(message, AIMessage):
                role = "assistant"
            else:
                continue

            message.content = cast(str, message.content)
            history.append(
                {
                    "role": role,
                    "content": message.content,
                }
            )

        return history



import sys

from prompt_toolkit.key_binding import KeyPressEvent
from rich.console import Console

from llm_chat_term import db, utils
from llm_chat_term.ui.prompt_menu import Menu


def handle_delete(event: KeyPressEvent, menu: Menu):
    if menu.selected_index == 0:
        return
    utils.delete_chat(menu.items[menu.selected_index])
    menu.items.pop(menu.selected_index)
    menu.selected_index = menu.selected_index - 1
    event.app.invalidate()


def handle_edit(event: KeyPressEvent, menu: Menu):
    utils.open_in_editor(menu.items[menu.selected_index])
    event.app.invalidate()


def select_chat() -> str:
    all_chats = db.list_all_chats()
    if len(all_chats) == 0:
        return create_new_chat()

    chats = ["Create new chat", *all_chats]

    menu = Menu(
        chats,
        " Select a chat (j/k to move, Enter to select, e to edit, d to delete, q to quit):\n",
    )

    menu.add_binding("e", handle_edit)

    menu.add_binding("d", handle_delete)

    result = menu.run()

    if result == 0:
        return create_new_chat()

    return chats[result]


def create_new_chat(*, allow_blank: bool = True) -> str:
    """Prompt the user to create a new chat."""
    from prompt_toolkit import prompt

    console = Console()
    if allow_blank:
        console.print("[green]Create a New Chat...[/green]")
    else:
        console.print("[green]Pick a name to save the chat before editing...[/green]")
    prompt_text = "Enter a name"
    prompt_text += "(blank for anonymous chat): " if allow_blank else ": "

    while True:
        try:
            chat_id = prompt(prompt_text)
        except KeyboardInterrupt:
            if allow_blank:
                console.print("\nExiting...")
                sys.exit(0)
            else:
                chat_id = ""
                console.print("Cannot edit conversation without a chat name...")

        chat_id = chat_id.strip()

        if chat_id:
            file_path = db.get_chat_file(chat_id)
            if file_path.exists():
                console.print(
                    (
                        f"[red]Chat [bold]{chat_id}[/bold] "
                        f"already exists at: [bold]{file_path}[/bold][/red]"
                    )
                )
                continue
            console.print(
                f"[blue]Chat conversation will be persisted to: {file_path}[/blue]"
            )

        return chat_id



import shlex
import subprocess

from llm_chat_term.llm.tools.definitions import CatFileCommand
from llm_chat_term.llm.tools.main import register_model


@register_model
def handle_cat(model: CatFileCommand):
    try:
        result = subprocess.run(  # noqa: S603
            ["cat", *shlex.split(model.arguments)],  # noqa: S607
            capture_output=True,
            text=True,
            check=False,
        )
    except Exception as e:
        return {"success": False, "exception": str(e)}

    return {
        "success": result.returncode == 0,
        "stdout": result.stdout,
        "stderr": result.stderr,
        "returncode": result.returncode,
        "command": f"cat {model.arguments}",
    }



import sys
from collections.abc import Callable

from prompt_toolkit import Application
from prompt_toolkit.key_binding import KeyBindings, KeyPressEvent
from prompt_toolkit.layout import FormattedTextControl, HSplit, Layout, Window


class Menu:
    def __init__(self, items: list[str], title: str):
        self.items = items
        self.title = title
        self.selected_index = 0
        self.result = 0
        self.kb = KeyBindings()
        self._setup_default_bindings()

    def _setup_default_bindings(self):
        @self.kb.add("j")
        @self.kb.add("down")
        def _(event: KeyPressEvent):
            self.selected_index = (self.selected_index + 1) % len(self.items)
            event.app.invalidate()

        @self.kb.add("k")
        @self.kb.add("up")
        def _(event: KeyPressEvent):
            self.selected_index = (self.selected_index - 1) % len(self.items)
            event.app.invalidate()

        @self.kb.add("enter")
        def _(event: KeyPressEvent):
            self.result = self.selected_index
            event.app.exit()

        @self.kb.add("q")
        @self.kb.add("c-c")
        def _(event: KeyPressEvent):
            event.app.exit()
            sys.exit(1)

    def add_binding(self, key: str, handler: Callable[[KeyPressEvent, "Menu"], None]):
        # Wrapper to add custom bindings
        @self.kb.add(key)
        def _(event: KeyPressEvent):
            handler(event, self)

    def run(self) -> int:
        # Create the layout
        layout = Layout(HSplit([Window(FormattedTextControl(self._get_menu_text))]))

        # Create and run the application
        app = Application(  # pyright: ignore[reportUnknownVariableType]
            layout=layout,
            key_bindings=self.kb,
            full_screen=False,
            mouse_support=True,
        )
        app.run()

        return self.result

    def _get_menu_text(self) -> list[tuple[str, str]]:
        result: list[tuple[str, str]] = []
        result.append(
            (
                "class:title",
                self.title,
            )
        )

        for i, item in enumerate(self.items):
            if i == self.selected_index:
                result.append(("green", f" > {item}\n"))
            else:
                result.append(("", f"   {item}\n"))

        return result



from rich.console import Console

from llm_chat_term.config import config

# fmt: off
COMMANDS: dict[str, list[str]] = {
    ":help": [
        "Displays this help message."
    ],
    ":info": [
        "Displays info about this chat."
    ],
    ":edit :e": [
        "Opens the conversation history in $EDITOR (vim by default).",
        "Edit it and save and it will be reloaded in the message history.",
        "You can edit the system prompt for the current conversation this way."
    ],
    ":chat": [
        "Display a menu to select a different chat."
    ],
    ":model": [
        "Display a menu to select a different chat."
    ],
    ":agent {on|off}": [
        "Enables/disables agent mode. Agent mode has access to tools that can",
        "affect your filesystem, use git etc.",
    ],
    ":redraw": [
        "Redraw the whole conversation."
    ],
    ":think {prompt}": [
        "Enable thinking mode only for this question (Claude only)."
    ],
    ":read {path}": [
        "Embed a text file in the prompt (replaces the line with :read)."
    ],
    ":web {url}": [
        "Embed a webpage contents in the prompt (replaces the line with :web)."
    ],
    ":exit": [
        "Exits the application. The conversation is saved if not anonymous chat."
    ],
}
# fmt: on


def print_help(console: Console):
    console.print("Available commands", style=f"bold {config.colors.system}")
    console.print()
    max_length: int = max([len(cmd) for cmd in COMMANDS])
    padding = 4
    for cmd, info_list in COMMANDS.items():
        console.print(cmd, style=config.colors.system, end="")
        d_length = max_length + padding - len(cmd)
        console.print(" " * d_length, end="")
        for idx, info in enumerate(info_list):
            if idx == 0:
                console.print(f"{info}", style=config.colors.system)
            else:
                console.print(
                    f"{' ' * (max_length + padding)}{info}",
                    style=config.colors.system,
                )
    console.print()
